@@ -34,6 +34,17 @@
 #ifndef QWAYLANDDISPLAY_H
 #define QWAYLANDDISPLAY_H
 
+//
+//  W A R N I N G
+//  -------------
+//
+// This file is not part of the Qt API.  It exists purely as an
+// implementation detail.  This header file may change from version to
+// version without notice, or even be removed.
+//
+// We mean it.
+//
+
 #include <QtCore/QObject>
 #include <QtCore/QRect>
 #include <QtCore/QPointer>
@@ -100,12 +111,14 @@ public:
     struct wl_surface *createSurface(void *handle);
     QWaylandShellSurface *createShellSurface(QWaylandWindow *window);
     struct ::wl_region *createRegion(const QRegion &qregion);
+    struct ::wl_subsurface *createSubSurface(QWaylandWindow *window, QWaylandWindow *parent);
 
     QWaylandClientBufferIntegration *clientBufferIntegration() const;
 
     QWaylandWindowManagerIntegration *windowManagerIntegration() const;
 
     void setCursor(struct wl_buffer *buffer, struct wl_cursor_image *image);
+    void setCursor(const QSharedPointer<QWaylandBuffer> &buffer, const QPoint &hotSpot);
 
     struct wl_display *wl_display() const { return mDisplay; }
     struct ::wl_registry *wl_registry() { return object(); }
@@ -127,8 +140,6 @@ public:
     QWaylandDataDeviceManager *dndSelectionHandler() const { return mDndSelectionHandler.data(); }
 
     QtWayland::qt_surface_extension *windowExtension() const { return mWindowExtension.data(); }
-    QtWayland::qt_sub_surface_extension *subSurfaceExtension() const { return mSubSurfaceExtension.data(); }
-    QtWayland::qt_output_extension *outputExtension() const { return mOutputExtension.data(); }
     QWaylandTouchExtension *touchExtension() const { return mTouchExtension.data(); }
     QtWayland::wl_text_input_manager *textInputManager() const { return mTextInputManager.data(); }
     QWaylandHardwareIntegration *hardwareIntegration() const { return mHardwareIntegration.data(); }
@@ -167,6 +178,7 @@ public slots:
 private:
     void waitForScreens();
     void exitWithError();
+    void checkError() const;
 
     struct Listener {
         RegistryListener listener;
@@ -176,8 +188,6 @@ private:
     struct wl_display *mDisplay;
     QtWayland::wl_compositor mCompositor;
     struct wl_shm *mShm;
-    QThread *mEventThread;
-    QWaylandEventThread *mEventThreadObject;
     QScopedPointer<QtWayland::wl_shell> mShell;
     QScopedPointer<QWaylandXdgShell> mShellXdg;
     QList<QWaylandScreen *> mScreens;
@@ -187,8 +197,7 @@ private:
     QWaylandInputDevice *mLastKeyboardFocusInputDevice;
     QScopedPointer<QWaylandDataDeviceManager> mDndSelectionHandler;
     QScopedPointer<QtWayland::qt_surface_extension> mWindowExtension;
-    QScopedPointer<QtWayland::qt_sub_surface_extension> mSubSurfaceExtension;
-    QScopedPointer<QtWayland::qt_output_extension> mOutputExtension;
+    QScopedPointer<QtWayland::wl_subcompositor> mSubCompositor;
     QScopedPointer<QWaylandTouchExtension> mTouchExtension;
     QScopedPointer<QWaylandQtKeyExtension> mQtKeyExtension;
     QScopedPointer<QWaylandWindowManagerIntegration> mWindowManagerIntegration;
