@@ -73,8 +73,7 @@ QWaylandWindow::QWaylandWindow(QWindow *window)
     , mWindowDecoration(0)
     , mMouseEventsInContentArea(false)
     , mMousePressedInContentArea(Qt::NoButton)
-    , m_cursorShape(Qt::ArrowCursor)
-    , mBuffer(0)
+    , m_cursor(Qt::ArrowCursor)
     , mWaitingForFrameSync(false)
     , mFrameCallback(0)
     , mRequestResizeSent(false)
@@ -86,21 +85,50 @@ QWaylandWindow::QWaylandWindow(QWindow *window)
     , mMask()
     , mBackingStore(Q_NULLPTR)
 {
-    init(mDisplay->createSurface(static_cast<QtWayland::wl_surface *>(this)));
-
     static WId id = 1;
     mWindowId = id++;
+    if (window->type() != Qt::Desktop)
+        initWindow();
+}
 
-    if (mDisplay->subSurfaceExtension())
-        mSubSurfaceWindow = new QWaylandSubSurface(this, mDisplay->subSurfaceExtension()->get_sub_surface_aware_surface(object()));
+QWaylandWindow::~QWaylandWindow()
+{
+    delete mWindowDecoration;
 
-    if (!(window->flags() & Qt::BypassWindowManagerHint)) {
+    if (isInitialized())
+        reset();
+
+    QList<QWaylandInputDevice *> inputDevices = mDisplay->inputDevices();
+    for (int i = 0; i < inputDevices.size(); ++i)
+        inputDevices.at(i)->handleWindowDestroyed(this);
+
+    const QWindow *parent = window();
+    foreach (QWindow *w, QGuiApplication::topLevelWindows()) {
+        if (w->transientParent() == parent)
+            QWindowSystemInterface::handleCloseEvent(w);
+    }
+
+    if (mMouseGrab == this) {
+        mMouseGrab = 0;
+    }
+}
+
+void QWaylandWindow::initWindow()
+{
+    init(mDisplay->createSurface(static_cast<QtWayland::wl_surface *>(this)));
+
+    if (shouldCreateSubSurface()) {
+        QWaylandWindow *p = static_cast<QWaylandWindow *>(QPlatformWindow::parent());
+        if (::wl_subsurface *ss = mDisplay->createSubSurface(this, p)) {
+            mSubSurfaceWindow = new QWaylandSubSurface(this, p, ss);
+        }
+    } else if (shouldCreateShellSurface()) {
         mShellSurface = mDisplay->createShellSurface(this);
     }
 
     if (mShellSurface) {
         // Set initial surface title
-        mShellSurface->setTitle(window->title());
+        mShellSurface->setTitle(window()->title());
 
         // The appId is the desktop entry identifier that should follow the
         // reverse DNS convention (see http://standards.freedesktop.org/desktop-entry-spec/latest/ar01s02.html),
@@ -122,14 +150,15 @@ QWaylandWindow::QWaylandWindow(QWindow *window)
         }
     }
 
-    if (QPlatformWindow::parent() && mSubSurfaceWindow) {
-        mSubSurfaceWindow->setParent(static_cast<const QWaylandWindow *>(QPlatformWindow::parent()));
-    } else if (window->transientParent() && mShellSurface) {
-        if (window->type() != Qt::Popup) {
-            mShellSurface->updateTransientParent(window->transientParent());
+    if (mShellSurface) {
+        if (window()->transientParent()) {
+            if (window()->type() != Qt::Popup) {
+                mShellSurface->updateTransientParent(window()->transientParent());
+            }
+        } else {
+            if (window()->type() != Qt::ToolTip)
+                mShellSurface->setTopLevel();
         }
-    } else if (mShellSurface && window->type() != Qt::ToolTip) {
-        mShellSurface->setTopLevel();
     }
 
     // Enable high-dpi rendering. Scale() returns the screen scale factor and will
@@ -138,38 +167,44 @@ QWaylandWindow::QWaylandWindow(QWindow *window)
     if (mDisplay->compositorVersion() >= 3)
         set_buffer_scale(scale());
 
-    setOrientationMask(window->screen()->orientationUpdateMask());
-    setWindowFlags(window->flags());
-    setGeometry_helper(window->geometry());
-    setMask(window->mask());
-    setWindowStateInternal(window->windowState());
-    handleContentOrientationChange(window->contentOrientation());
+    if (QScreen *s = window()->screen())
+        setOrientationMask(s->orientationUpdateMask());
+    setWindowFlags(window()->flags());
+    setGeometry_helper(window()->geometry());
+    setMask(window()->mask());
+    setWindowStateInternal(window()->windowState());
+    handleContentOrientationChange(window()->contentOrientation());
 }
 
-QWaylandWindow::~QWaylandWindow()
+bool QWaylandWindow::shouldCreateShellSurface() const
 {
-    delete mWindowDecoration;
+    if (shouldCreateSubSurface())
+        return false;
+
+    if (window()->inherits("QShapedPixmapWindow"))
+        return false;
+
+    if (qEnvironmentVariableIsSet("QT_WAYLAND_USE_BYPASSWINDOWMANAGERHINT"))
+        return !(window()->flags() & Qt::BypassWindowManagerHint);
+
+    return true;
+}
+
+bool QWaylandWindow::shouldCreateSubSurface() const
+{
+    return QPlatformWindow::parent() != Q_NULLPTR;
+}
+
+void QWaylandWindow::reset()
+{
+    delete mShellSurface;
+    mShellSurface = 0;
+    delete mSubSurfaceWindow;
+    mSubSurfaceWindow = 0;
+    destroy();
 
-    if (isInitialized()) {
-        delete mShellSurface;
-        destroy();
-    }
     if (mFrameCallback)
         wl_callback_destroy(mFrameCallback);
-
-    QList<QWaylandInputDevice *> inputDevices = mDisplay->inputDevices();
-    for (int i = 0; i < inputDevices.size(); ++i)
-        inputDevices.at(i)->handleWindowDestroyed(this);
-
-    const QWindow *parent = window();
-    foreach (QWindow *w, QGuiApplication::topLevelWindows()) {
-        if (w->transientParent() == parent)
-            QWindowSystemInterface::handleCloseEvent(w);
-    }
-
-    if (mMouseGrab == this) {
-        mMouseGrab = 0;
-    }
 }
 
 QWaylandWindow *QWaylandWindow::fromWlSurface(::wl_surface *surface)
@@ -184,9 +219,17 @@ WId QWaylandWindow::winId() const
 
 void QWaylandWindow::setParent(const QPlatformWindow *parent)
 {
-    const QWaylandWindow *parentWaylandWindow = static_cast<const QWaylandWindow *>(parent);
-    if (subSurfaceWindow()) {
-        subSurfaceWindow()->setParent(parentWaylandWindow);
+    QWaylandWindow *oldparent = mSubSurfaceWindow ? mSubSurfaceWindow->parent() : 0;
+    if (oldparent == parent)
+        return;
+
+    if (mSubSurfaceWindow && parent) { // new parent, but we were a subsurface already
+        delete mSubSurfaceWindow;
+        QWaylandWindow *p = const_cast<QWaylandWindow *>(static_cast<const QWaylandWindow *>(parent));
+        mSubSurfaceWindow = new QWaylandSubSurface(this, p, mDisplay->createSubSurface(this, p));
+    } else { // we're changing role, need to make a new wl_surface
+        reset();
+        initWindow();
     }
 }
 
@@ -214,7 +257,10 @@ void QWaylandWindow::setGeometry_helper(const QRect &rect)
                 qBound(window()->minimumWidth(), rect.width(), window()->maximumWidth()),
                 qBound(window()->minimumHeight(), rect.height(), window()->maximumHeight())));
 
-    if (shellSurface() && window()->transientParent() && window()->type() != Qt::Popup)
+    if (mSubSurfaceWindow) {
+        QMargins m = QPlatformWindow::parent()->frameMargins();
+        mSubSurfaceWindow->set_position(rect.x() + m.left(), rect.y() + m.top());
+    } else if (shellSurface() && window()->transientParent() && window()->type() != Qt::Popup)
         shellSurface()->updateTransientParent(window()->transientParent());
 }
 
@@ -384,9 +430,8 @@ void QWaylandWindow::requestResize()
 
 void QWaylandWindow::attach(QWaylandBuffer *buffer, int x, int y)
 {
-    mBuffer = buffer;
-    if (mBuffer)
-        attach(mBuffer->buffer(), x, y);
+    if (buffer)
+        attach(buffer->buffer(), x, y);
     else
         QtWayland::wl_surface::attach(0, 0, 0);
 }
@@ -397,11 +442,6 @@ void QWaylandWindow::attachOffset(QWaylandBuffer *buffer)
     mOffset = QPoint();
 }
 
-QWaylandBuffer *QWaylandWindow::attached() const
-{
-    return mBuffer;
-}
-
 void QWaylandWindow::damage(const QRect &rect)
 {
     //We have to do sync stuff before calling damage, or we might
@@ -411,9 +451,7 @@ void QWaylandWindow::damage(const QRect &rect)
         wl_callback_add_listener(mFrameCallback,&QWaylandWindow::callbackListener,this);
         mWaitingForFrameSync = true;
     }
-    if (mBuffer) {
-        damage(rect.x(), rect.y(), rect.width(), rect.height());
-    }
+    damage(rect.x(), rect.y(), rect.width(), rect.height());
 }
 
 const wl_callback_listener QWaylandWindow::callbackListener = {
@@ -547,7 +585,10 @@ bool QWaylandWindow::createDecoration()
         decoration = false;
     if (window()->flags() & Qt::BypassWindowManagerHint)
         decoration = false;
+    if (mSubSurfaceWindow)
+        decoration = false;
 
+    bool hadDecoration = mWindowDecoration;
     if (decoration && !decorationPluginFailed) {
         if (!mWindowDecoration) {
             QStringList decorations = QWaylandDecorationFactory::keys();
@@ -578,15 +619,20 @@ bool QWaylandWindow::createDecoration()
                 return false;
             }
             mWindowDecoration->setWaylandWindow(this);
-            if (subSurfaceWindow()) {
-                subSurfaceWindow()->adjustPositionOfChildren();
-            }
         }
     } else {
         delete mWindowDecoration;
         mWindowDecoration = 0;
     }
 
+    if (hadDecoration != (bool)mWindowDecoration) {
+        foreach (QWaylandSubSurface *subsurf, mChildren) {
+            QPoint pos = subsurf->window()->geometry().topLeft();
+            QMargins m = frameMargins();
+            subsurf->set_position(pos.x() + m.left(), pos.y() + m.top());
+        }
+    }
+
     return mWindowDecoration;
 }
 
@@ -705,17 +751,17 @@ void QWaylandWindow::handleMouseEventWithDecoration(QWaylandInputDevice *inputDe
     }
 }
 
-void QWaylandWindow::setMouseCursor(QWaylandInputDevice *device, Qt::CursorShape shape)
+void QWaylandWindow::setMouseCursor(QWaylandInputDevice *device, const QCursor &cursor)
 {
-    if (m_cursorShape != shape || device->serial() > device->cursorSerial()) {
-        device->setCursor(shape, mScreen);
-        m_cursorShape = shape;
+    if (device->serial() >= device->cursorSerial()) {
+        device->setCursor(cursor, mScreen);
+        m_cursor = cursor;
     }
 }
 
 void QWaylandWindow::restoreMouseCursor(QWaylandInputDevice *device)
 {
-    setMouseCursor(device, window()->cursor().shape());
+    setMouseCursor(device, window()->cursor());
 }
 
 void QWaylandWindow::requestActivateWindow()
